package minesweeper;

import java.awt.Color;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.Frame;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.Window;
import java.util.Date;
import javax.swing.BorderFactory;
import javax.swing.border.BevelBorder;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.Timer;
import minesweeper.event.MineFieldEvent;
import minesweeper.event.MineFieldListener;

/**
 *
 * The <code>MineSweeper</code> main frame.
 *
 */
public class MainFrame
  extends javax.swing.JFrame
{
  /** The canvas background color. */
  private static final Color CANVASBACKGROUNDCOLOR= Color.gray;

  /** A smiling smiley. */
  private static final ImageIcon HAPPYSMILEY= new ImageIcon( "minesweeper/images/HappySmiley.gif" );

  /** A sad smiley. */
  private static final ImageIcon SADSMILEY= new ImageIcon( "minesweeper/images/SadSmiley.gif" );

  /** A wondering smiley. */
  private static final ImageIcon WONDERINGSMILEY= new ImageIcon( "minesweeper/images/WonderingSmiley.gif" );

  /** The text field displaying the number of remaining mines. */
  private final JTextField tfMineCount= new JTextField( 5 );

  /** The text field displaying the elapsed time. */
  private final JTextField tfTimer= new JTextField( 5 );

  /** The reset button. */
  private final JButton btnReset= new JButton( HAPPYSMILEY )
                                      {
                                        /**
                                         * Indicates that this button cannot be traversed.
                                         */
                                        public boolean isFocusTraversable()
                                        {
                                          return false;
                                        }
                                      };

  /** The mine field. */
  private MineField mfMineField;

  /** The initial number of mines. */
  private int iInitialMines;

  /** The number of remaining mines. */
  private int iRemainingMines;

  /** The flag indicating whether the game is running. */
  private boolean bGameAllowed= true;

  /** The time used to update second display. */
  private final Timer tTimer= new Timer( 1000, new ActionListener()
                                                   {
                                                     public void actionPerformed( final ActionEvent ae )
                                                     {
                                                       final long     lTime= (long)((new Date()).getTime() / 1000);

                                                       tfTimer.setText( Long.toString( lTime - lStartTime ) );
                                                     }
                                                   } );

  /** The date (in seconds since Epoch) at which the current play started. */
  private long lStartTime= 0;


  /**************************************** constructors ****************************************/

  /**
   * Creates a new <code>MainFrame</code> object.
   *
   * @pre      0 < iColumns
   * @pre      0 < iRows
   * @pre      0 <= iMines
   */
  public MainFrame( final int iColumns, final int iRows, final int iMines )
  {
    super( "Minesweeper" );

    iInitialMines= iMines;

    final Container     cntContentPane= getContentPane();

    /***** setting component properties *****/
    tfMineCount.setHorizontalAlignment( JTextField.CENTER );
    tfMineCount.setEnabled( false );
    tfMineCount.setDisabledTextColor( Color.red );
    tfTimer.setHorizontalAlignment( JTextField.CENTER );
    tfTimer.setEnabled( false );
    tfTimer.setDisabledTextColor( Color.red );

    /***** building the mine field *****/
    mfMineField= new MineField( iColumns, iRows, iMines );
    mfMineField.addListener( new MineFieldListener()
                                 {
                                   /**
                                    * Invoked when the player adds a new flag.
                                    *
                                    * @pre      mfce != null
                                    */
                                   public void flagAdded( final MineFieldEvent mfe )
                                   {
                                     --iRemainingMines;
                                     refreshMineCountField();
                                   }

                                   /**
                                    * Invoked when the player removes a flag.
                                    *
                                    * @pre      mfce != null
                                    */
                                   public void flagRemoved( final MineFieldEvent mfe )
                                   {
                                     ++iRemainingMines;
                                     refreshMineCountField();
                                   }

                                   /**
                                    * Invoked when the game starts.
                                    *
                                    * @pre      mfe != null
                                    */
                                   public void gameStarts( final MineFieldEvent mfe )
                                   {
                                     lStartTime= (long)((new Date()).getTime() / 1000);
                                     tTimer.start();
                                   }

                                   /**
                                    * Invoked when the player stepped on a mine.
                                    *
                                    * @pre      mfce != null
                                    */
                                   public void steppedOnAMine( final MineFieldEvent mfe )
                                   {
                                     gameOver();
                                   }
                                 } );

    /***** adding a listener to the reset button *****/
    btnReset.addActionListener( new ActionListener()
                                    {
                                      /**
                                       * Invoked when an action occurs.
                                       */
                                      public void actionPerformed( final ActionEvent ae )
                                      {
                                        reset( false );
                                      }
                                    } );

    /***** building the top pane *****/
    final JPanel                   pTopPanel= new JPanel( new GridBagLayout() );
    final GridBagConstraints       gbcConstraints= new GridBagConstraints();

    pTopPanel.setBorder( BorderFactory.createBevelBorder( BevelBorder.LOWERED ) );
    gbcConstraints.gridx= gbcConstraints.gridy= GridBagConstraints.RELATIVE;
    gbcConstraints.gridwidth= 1;
    gbcConstraints.gridheight= GridBagConstraints.REMAINDER;
    gbcConstraints.weightx= gbcConstraints.weighty= 1;
    gbcConstraints.anchor= GridBagConstraints.WEST;
    gbcConstraints.fill= GridBagConstraints.NONE;
    gbcConstraints.insets= new Insets( 5, 5, 5, 5 );
    pTopPanel.add( tfMineCount, gbcConstraints );
    gbcConstraints.anchor= GridBagConstraints.CENTER;
    pTopPanel.add( btnReset, gbcConstraints );
    gbcConstraints.anchor= GridBagConstraints.EAST;
    pTopPanel.add( tfTimer, gbcConstraints );

    /***** building the bottom pane *****/
    final JPanel    pBottomPanel= new JPanel( new GridBagLayout() );

    pBottomPanel.setBorder( BorderFactory.createBevelBorder( BevelBorder.LOWERED ) );
    gbcConstraints.gridx= gbcConstraints.gridy= 0;
    gbcConstraints.anchor= GridBagConstraints.CENTER;
    gbcConstraints.fill= GridBagConstraints.NONE;
    gbcConstraints.insets= new Insets( 2, 2, 2, 2 );
    gbcConstraints.ipadx= gbcConstraints.ipady= 0;
    pBottomPanel.add( mfMineField, gbcConstraints );

    /***** building the main pane *****/
    final JPanel    pMainPanel= new JPanel( new GridBagLayout() );

    pMainPanel.setBorder( BorderFactory.createBevelBorder( BevelBorder.RAISED ) );
    gbcConstraints.gridx= gbcConstraints.gridy= 0;
    gbcConstraints.gridwidth= gbcConstraints.gridheight= 1;
    gbcConstraints.anchor= GridBagConstraints.CENTER;
    gbcConstraints.fill= GridBagConstraints.BOTH;
    gbcConstraints.insets= new Insets( 5, 5, 5, 5 );
    pMainPanel.add( pTopPanel, gbcConstraints );
    gbcConstraints.gridy= 1;
    pMainPanel.add( pBottomPanel, gbcConstraints );

    cntContentPane.add( pMainPanel );

    /***** adding a key listener to handle the F2 key *****/
    addKeyListener( new KeyAdapter()
                        {
                          public void keyPressed( final KeyEvent ke )
                          {
                            /***** why dos F2 is received only with a modifier ????? *****/
                            if (ke.getKeyCode() == KeyEvent.VK_F2)
                              reset( false );
                          }
                        } );

    /***** reset everything *****/
    reset( true );

    /***** finalizing frame creation *****/
    pack();

    /***** centering the window *****/
    final Dimension     dimWindow= getSize();
    final Dimension     dimScreen= java.awt.Toolkit.getDefaultToolkit().getScreenSize();

    setResizable( false );
    setLocation( (dimScreen.width - dimWindow.width) / 2,
                 (dimScreen.height - dimWindow.height) / 2 );

    setVisible( true );
  }


  /**************************************** miscellaneous ****************************************/

  /**
   * Disables the mine field and reveals wrong mines.
   */
  protected void gameOver()
  {
    bGameAllowed= false;
    tTimer.stop();
    lStartTime= 0;
    mfMineField.stopGame();
    btnReset.setIcon( SADSMILEY );
  }

  /**
   * Resets the game.
   *
   * @param    bFirstCall     indicates whether this is an initialisation and not a reset 
   */
  protected void reset( final boolean bFirstCall )
  {
    iRemainingMines= iInitialMines;
    refreshMineCountField();
    tfTimer.setText( "0" );
    if (!bFirstCall)
    {
      mfMineField.reset( bFirstCall );
      mfMineField.repaint();
      btnReset.setIcon( HAPPYSMILEY );
    }
    mfMineField.transferFocus();
    bGameAllowed= true;
  }

  /**
   * Refreshes the text field that displays the number of remaining mines.
   */
  protected void refreshMineCountField()
  {
    tfMineCount.setText( Integer.toString( iRemainingMines ) );
  }
}
