package minesweeper;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.InputEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Point;
import javax.swing.event.EventListenerList;
import javax.swing.JTable;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableCellRenderer;
import javax.swing.table.TableColumnModel;
import javax.swing.table.TableColumn;
import javax.swing.table.TableModel;
import minesweeper.event.MineFieldCellEvent;
import minesweeper.event.MineFieldCellListener;
import minesweeper.event.MineFieldEvent;
import minesweeper.event.MineFieldListener;

/**
 *
 * A mine fied.
 *
 * @inv        0 < iColumns
 * @inv        0 < iRows
 * @inv        mfcCells != null
 * @inv        mfcCells.length() == (iColumns*iRows)
 *
 */
public class MineField
  extends javax.swing.JPanel
{
    /**
     *
     * A mine field cell listener with cell coordinates.
     *
     */
    private class MineFieldCellListenerWithCoordinates
      implements minesweeper.event.MineFieldCellListener
    {
      /** The column of the listened cell. */
      private int iColumn;

      /** The row of the listened cell. */
      private int iRow;

      /**
       * Build a new listener with the specified coordinates.
       */
      public MineFieldCellListenerWithCoordinates( final int iRow, final int iColumn )
      {
        this.iRow= iRow;
        this.iColumn= iColumn;
      }

      /**
       * Invoked when the display of one of the cell has changed.
       *
       * @pre     mfce != null
       */
      public void displayChanged( final MineFieldCellEvent mfce )
      {
        tmTableModel.fireTableCellUpdated( iRow, iColumn );
      }

      /**
       * Invoked when the player adds a new flag.
       *
       * @pre      mfce != null
       */
      public void flagAdded( final MineFieldCellEvent mfce )
      {
        fireFlagAdded();
      }

      /**
       * Invoked when the player removes a flag.
       *
       * @pre      mfce != null
       */
      public void flagRemoved( final MineFieldCellEvent mfce )
      {
        fireFlagRemoved();
      }

      /**
       * Invoked when the player stepped on a mine.
       *
       * @pre      mfce != null
       */
      public void steppedOnAMine( final MineFieldCellEvent mfce )
      {
        fireSteppedOnAMine();
      }
    }

  /** The maximum width of columns. */
  private static final int CELLWIDTH= 16;

  /** The number of columns. */
  private int iColumns;

  /** The number of rows. */
  private int iRows;

  /** The initial number of mines. */
  private int iInitialMines;

  /** The flag indicating whether the game is running. */
  private boolean bGameAllowed= true;

  /** The flag indicating whether the game is started. */
  private boolean bGameStarted= false;

  /** The array of mine field cells. */
  private MineFieldCell[][] mfcCells;

  /** The JTable. */
  private JTable jtTable;

  /** The table model. */
  private final AbstractTableModel tmTableModel= new AbstractTableModel()
                                                     {
                                                       /**
                                                        * Returns the number of columns in the model.
                                                        */
                                                       public int getColumnCount()
                                                       {
                                                         return iColumns;
                                                       }

                                                       /**
                                                        * Returns the number of rows in the model.
                                                        */
                                                       public int getRowCount()
                                                       {
                                                         return iRows;
                                                       }

                                                       /**
                                                        * Returns the value for the cell at columnIndex and rowIndex.
                                                        */
                                                       public Object getValueAt( final int iRow, final int iColumn )
                                                       {
                                                         //System.err.println( "getValueAt( " + iRow +", " + iColumn + " )" );
                                                         return getCell( iRow, iColumn );
                                                       }
                                                     };


  /**************************************** constructors ****************************************/

  /**
   * Creates a new mine field, with the specified number of columns and rows and mines.
   *
   * @pre      0 < iColumns
   * @pre      0 < iRows
   * @pre      0 <= iMines
   */
  public MineField( final int iColumns, final int iRows, final int iMines )
  {
    super( new GridBagLayout() );
    iInitialMines= iMines;

    /***** setting the dimension of this field *****/
    setDimension( iColumns, iRows );

    /***** creating the table *****/
    jtTable= new JTable( tmTableModel );
    jtTable.setIntercellSpacing( new Dimension( 0, 0 ) );
    jtTable.setEnabled( false );
    jtTable.setAutoResizeMode( JTable.AUTO_RESIZE_OFF );

    /***** setting the maximum size of the columns and the renderer ******/
    final TableColumnModel    tcmModel= jtTable.getColumnModel();
    final TableCellRenderer   tcrRenderer= new TableCellRenderer()
                                               {
                                                 /**
                                                  * Returns the component used for drawing the cell.
                                                  * This method is used to configure the renderer appropriately before drawing.
                                                  */
                                                 public Component getTableCellRendererComponent( final JTable jtTable,
                                                                                                 final Object oValue,
                                                                                                 final boolean bIsSelected,
                                                                                                 final boolean bhasFocus,
                                                                                                 final int iRow,
                                                                                                 final int iColumn )
                                                 {
                                                   return ((MineFieldCell)oValue).getComponent();
                                                 }
                                               };

    for ( int iColumnCounter= 0 ; iColumnCounter < tcmModel.getColumnCount() ; ++iColumnCounter )
    {
      final TableColumn       tcCurrentColumn= tcmModel.getColumn( iColumnCounter );

      tcCurrentColumn.setCellRenderer( tcrRenderer );
      tcCurrentColumn.setMinWidth( CELLWIDTH );
      tcCurrentColumn.setPreferredWidth( CELLWIDTH );
      tcCurrentColumn.setMaxWidth( CELLWIDTH );
    }

    /***** adding a mouse listener *****/
    jtTable.addMouseListener( new MouseAdapter()
                                  {
                                    /**
				     * Invoked when the mouse has been clicked on the table.
				     */
                                    public void mouseClicked( final MouseEvent me )
                                    {
                                      if (   bGameAllowed
					  && ((me.getModifiers() & InputEvent.BUTTON1_MASK) != 0)
                                          && ((me.getModifiers() & InputEvent.BUTTON3_MASK) != 0))
                                      {
                                        startGame( me.getPoint() );
                                        revealSurrounders( me.getPoint() );
                                      }
                                    }

                                    /**
				     * Invoked when a mouse button has been pressed on the table.
				     */
                                    public void mousePressed( final MouseEvent me )
                                    {
                                      if (bGameAllowed && (me.getModifiers() & InputEvent.BUTTON3_MASK) != 0)
                                      {
                                        startGame( me.getPoint() );
                                        toggleFlag( me.getPoint() );
                                      }
                                    }

                                    /**
				     * Invoked when the mouse has been released on the table.
				     */
                                    public void mouseReleased( final MouseEvent me )
                                    {
                                      if (   bGameAllowed
                                          && (   ((me.getModifiers() & InputEvent.BUTTON1_MASK) != 0)
                                              || ((me.getModifiers() & InputEvent.BUTTON3_MASK) != 0)))
                                      {
                                        startGame( me.getPoint() );
                                        if (   ((me.getModifiers() & InputEvent.BUTTON1_MASK) != 0)
                                            && ((me.getModifiers() & InputEvent.BUTTON3_MASK) != 0))
                                          revealSurrounders( me.getPoint() );
                                        else
                                          if ((me.getModifiers() & InputEvent.BUTTON1_MASK) != 0)
                                            reveal( me.getPoint() );
                                      }
                                    }
                                  } );

    /***** reset everything *****/
    reset( true );

    /***** adding the table *****/
    final GridBagConstraints       gbcConstraints= new GridBagConstraints();

    gbcConstraints.gridx= 0;
    gbcConstraints.gridy= 0;
    gbcConstraints.gridwidth= GridBagConstraints.REMAINDER;
    gbcConstraints.gridheight= GridBagConstraints.REMAINDER;
    gbcConstraints.weightx= 0;
    gbcConstraints.weighty= 0;
    gbcConstraints.anchor= GridBagConstraints.CENTER;
    gbcConstraints.fill= GridBagConstraints.NONE;
    add( jtTable, gbcConstraints );
  }


  /**************************************** getters ****************************************/

  /**
   * Returns the cell of specified position
   *
   * @pre      (0 <= iRow) && (iRow < getRowCount())
   * @pre      (0 <= iColumn) && (iColumn < getColumnCount())
   * @post     return != null
   */
  public MineFieldCell getCell( final int iRow, final int iColumn )
  {
    return mfcCells[iRow][iColumn];
  }

  /**
   * Returns the number of columns.
   *
   * @post     0 < return
   */
  public int getColumnCount()
  {
    return iColumns;
  }

  /**
   * Returns the number of rows.
   *
   * @post     0 < return
   */
  public int getRowCount()
  {
    return iRows;
  }


  /**************************************** setters ****************************************/

  /**
   * Dispatches the mines in this field, avoiding the cell of specified coordinates.
   */
  public void dispatchMines( final int iAvoidRow, final int iAvoidColumn )
  {
    for ( int iMineCounter= 0 ; iMineCounter < iInitialMines ; ++iMineCounter )
    {
      boolean  bOk= false;

      do
      {
	final int             iColumn= (int)(iColumns*Math.random());
	final int             iRow= (int)(iRows*Math.random());
        final MineFieldCell   mfc= getCell( iRow, iColumn );

        if (   !(   (iAvoidColumn-1 <= iColumn) && (iColumn <= iAvoidColumn+1)
                 && (iAvoidRow-1 <= iRow) && (iRow <= iAvoidRow+1))
            && !mfc.doesContainAMine())
	{
          mfc.settleAMine();
          bOk= true;
	}
      }
      while (!bOk);
    }
  }

  /**
   * Reset the mine field.
   *
   * @param    bFirstCall     indicates whether this is an initialisation and not a reset 
   */
  public void reset( final boolean bFirstCall )
  {
    if (!bFirstCall)
      for ( int iColumnCounter= 0 ; iColumnCounter < getColumnCount() ; ++iColumnCounter )
        for ( int iRowCounter= 0 ; iRowCounter < getRowCount() ; ++iRowCounter )
          getCell( iRowCounter, iColumnCounter ).reset();
    bGameAllowed= true;
    bGameStarted= false;
  }

  /**
   * Sets the number of columns and rows of this mine field.
   *
   * @pre      0 < iColumns
   * @pre      0 < iRows
   */
  public void setDimension( final int iColumns, final int iRows )
  {
    this.iColumns= iColumns;
    this.iRows= iRows;

    /***** creating and filling the array of cells *****/
    mfcCells= new MineFieldCell[getRowCount()][getColumnCount()];
    for ( int iColumnCounter= 0 ; iColumnCounter < getColumnCount() ; ++iColumnCounter )
      for ( int iRowCounter= 0 ; iRowCounter < getRowCount() ; ++iRowCounter )
      {
	final MineFieldCell   mfcNewCell= new MineFieldCell();

        mfcCells[iRowCounter][iColumnCounter]= mfcNewCell;
        mfcNewCell.addListener( new MineFieldCellListenerWithCoordinates( iRowCounter, iColumnCounter ) );
      }

    /***** linking cells *****/
    for ( int iColumnCounter= 0 ; iColumnCounter < getColumnCount() ; ++iColumnCounter )
    {
      final boolean      bFirstColumn= (iColumnCounter == 0);
      final boolean      bLastColumn= (iColumnCounter == (getColumnCount() - 1));

      for ( int iRowCounter= 0 ; iRowCounter < getRowCount() ; ++iRowCounter )
      {
        final boolean         bFirstRow= (iRowCounter == 0);
        final boolean         bLastRow= (iRowCounter == (getRowCount() -1));
        final MineFieldCell   mfcCurrent= getCell( iRowCounter, iColumnCounter );

        if (!bFirstColumn)
        {
          getCell( iRowCounter, iColumnCounter-1 ).registerAsSurrounder( mfcCurrent );
          if (!bFirstRow)
            getCell( iRowCounter-1, iColumnCounter-1 ).registerAsSurrounder( mfcCurrent );
          if (!bLastRow)
            getCell( iRowCounter+1, iColumnCounter-1 ).registerAsSurrounder( mfcCurrent );
	}
        if (!bLastColumn)
	{
          getCell( iRowCounter, iColumnCounter+1 ).registerAsSurrounder( mfcCurrent );
          if (!bFirstRow)
            getCell( iRowCounter-1, iColumnCounter+1 ).registerAsSurrounder( mfcCurrent );
          if (!bLastRow)
            getCell( iRowCounter+1, iColumnCounter+1 ).registerAsSurrounder( mfcCurrent );
	}
        if (!bFirstRow)
          getCell( iRowCounter-1, iColumnCounter ).registerAsSurrounder( mfcCurrent );
        if (!bLastRow)
          getCell( iRowCounter+1, iColumnCounter ).registerAsSurrounder( mfcCurrent );
      }
    }
  }

  /**
   * Starts the game: dispatches the mines, avoiding the specified point.
   */
  public void startGame( final Point ptNoMineThere )
  {
    if (!bGameStarted)
    {
      dispatchMines( jtTable.rowAtPoint( ptNoMineThere ), jtTable.columnAtPoint( ptNoMineThere ) );
      bGameStarted= true;
      fireGameStarts();
    }
  }

  /**
   * Stops the game.
   */
  public void stopGame()
  {
    bGameAllowed= false;
    bGameStarted= false;
    for ( int iColumnCounter= 0 ; iColumnCounter < getColumnCount() ; ++iColumnCounter )
      for ( int iRowCounter= 0 ; iRowCounter < getRowCount() ; ++iRowCounter )
        getCell( iRowCounter, iColumnCounter ).revealError();
  }


  /**************************************** mouse event answers ****************************************/

  /**
   * Reveals the cell of specified location.
   */
  protected void reveal( final Point ptLocation )
  {
    final int  iColumn= jtTable.columnAtPoint( ptLocation );
    final int  iRow= jtTable.rowAtPoint( ptLocation );

    if ((0 <= iColumn) && (0 <= iRow))
      getCell( iRow, iColumn ).reveal();
  }

  /**
   * Reveals the surrounders of the cell of specified location.
   */
  protected void revealSurrounders( final Point ptLocation )
  {
    final int  iColumn= jtTable.columnAtPoint( ptLocation );
    final int  iRow= jtTable.rowAtPoint( ptLocation );

    if ((0 <= iColumn) && (0 <= iRow))
      getCell( iRow, iColumn ).revealSurrounders();
  }

  /**
   * Toggles a flag at the cell of specified location.
   */
  protected void toggleFlag( final Point ptLocation )
  {
    final int  iColumn= jtTable.columnAtPoint( ptLocation );
    final int  iRow= jtTable.rowAtPoint( ptLocation );

    if ((0 <= iColumn) && (0 <= iRow))
      getCell( iRow, iColumn ).toggleFlag();
  }


  /**************************************** listener management ****************************************/

  /** The source of events. */
  private MineFieldEvent mfeEvent= null;

  /**
   * Adds the specified mine field listener to receive event from this field.
   *
   * @pre      mflListener != null
   */
  public void addListener( final MineFieldListener mflListener )
  {
    listenerList.add( MineFieldListener.class, mflListener );
  }

  /**
   * Removes the specified mine field listener so that it no longer receive events from this field.
   *
   * @pre      mflListener != null
   */
  public void removeListener( final MineFieldListener mflListener )
  {
    listenerList.remove( MineFieldListener.class, mflListener );
  }

  /**
   * Fires a flagAdded event.
   */
  protected void fireFlagAdded()
  {
    final Object[]  oListeners= listenerList.getListenerList();

    for ( int iListenerCounter= oListeners.length-2 ;
          iListenerCounter >= 0 ;
          iListenerCounter-= 2 )
      if (oListeners[iListenerCounter] == MineFieldListener.class)
      {
        if (mfeEvent == null)
          mfeEvent= new MineFieldEvent( this );
        ((MineFieldListener)oListeners[iListenerCounter+1]).flagAdded( mfeEvent );
      }
  }

  /**
   * Fires a flagRemoved event.
   */
  protected void fireFlagRemoved()
  {
    final Object[]  oListeners= listenerList.getListenerList();

    for ( int iListenerCounter= oListeners.length-2 ;
          iListenerCounter >= 0 ;
          iListenerCounter-= 2 )
      if (oListeners[iListenerCounter] == MineFieldListener.class)
      {
        if (mfeEvent == null)
          mfeEvent= new MineFieldEvent( this );
        ((MineFieldListener)oListeners[iListenerCounter+1]).flagRemoved( mfeEvent );
      }
  }

  /**
   * Fires a gameStarts event.
   */
  protected void fireGameStarts()
  {
    final Object[]  oListeners= listenerList.getListenerList();

    for ( int iListenerCounter= oListeners.length-2 ;
          iListenerCounter >= 0 ;
          iListenerCounter-= 2 )
      if (oListeners[iListenerCounter] == MineFieldListener.class)
      {
        if (mfeEvent == null)
          mfeEvent= new MineFieldEvent( this );
        ((MineFieldListener)oListeners[iListenerCounter+1]).gameStarts( mfeEvent );
      }
  }

  /**
   * Fires a steppedOnAMine event.
   */
  protected void fireSteppedOnAMine()
  {
    final Object[]  oListeners= listenerList.getListenerList();

    for ( int iListenerCounter= oListeners.length-2 ;
          iListenerCounter >= 0 ;
          iListenerCounter-= 2 )
      if (oListeners[iListenerCounter] == MineFieldListener.class)
      {
        if (mfeEvent == null)
          mfeEvent= new MineFieldEvent( this );
        ((MineFieldListener)oListeners[iListenerCounter+1]).steppedOnAMine( mfeEvent );
      }
  }
}
