package minesweeper;

import java.awt.Color;
import java.awt.Component;
import java.util.LinkedList;
import java.util.ListIterator;
import javax.swing.event.EventListenerList;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.SwingConstants;
import minesweeper.event.MineFieldCellEvent;
import minesweeper.event.MineFieldCellListener;

/**
 *
 * A cell of the mine field.
 *
 * @invariant  (0 <= iSurroundingMines) && (iSurroundingMines <= 8)
 * @invariant  forall Object o in lstSurrounders | o instanceof MineFieldCell
 *
 */
public class MineFieldCell
{
  /** An empty cell. */
  private static final JLabel EMPTY= new JLabel();

  /** A mine that was wrongly supposed. */
  private static final JLabel ERROR= new JLabel( new ImageIcon( "minesweeper/images/Error.gif" ) );

  /** A mine. */
  private static final JLabel MINE= new JLabel( new ImageIcon( "minesweeper/images/Mine.gif" ) );

  /** A flag. */
  private static final ImageIcon FLAG= new ImageIcon( "minesweeper/images/Flag.gif" );

  /** An array of digits; these labels should not be modified. */
  private static final JLabel[] DIGITS= { new JLabel( "1", SwingConstants.CENTER ),
                                          new JLabel( "2", SwingConstants.CENTER ),
                                          new JLabel( "3", SwingConstants.CENTER ),
                                          new JLabel( "4", SwingConstants.CENTER ),
                                          new JLabel( "5", SwingConstants.CENTER ),
                                          new JLabel( "6", SwingConstants.CENTER ),
                                          new JLabel( "7", SwingConstants.CENTER ),
                                          new JLabel( "8", SwingConstants.CENTER )
                                        };

  /**
   *
   * The "static constructor" of this class.
   *
   */
  static
  {
    DIGITS[0].setForeground( new Color( 0, 0, 255 ) );
    DIGITS[1].setForeground( new Color( 0, 128, 0 ) );
    DIGITS[2].setForeground( new Color( 255, 0, 0 ) );
    DIGITS[3].setForeground( new Color( 0, 0, 128 ) );
    DIGITS[4].setForeground( new Color( 128, 0, 0 ) );
    DIGITS[5].setForeground( new Color( 0, 255, 0 ) );
    DIGITS[6].setForeground( new Color( 255, 255, 0 ) );
    DIGITS[7].setForeground( new Color( 0, 255, 255 ) );
    for ( int iLabelCounter= 0 ; iLabelCounter < 8 ; ++iLabelCounter )
    {
      DIGITS[iLabelCounter].setBackground( Color.lightGray );
      DIGITS[iLabelCounter].setOpaque( true );
    }
    EMPTY.setBackground( Color.lightGray );
    EMPTY.setOpaque( true );
    ERROR.setBackground( Color.lightGray );
    ERROR.setOpaque( true );
    MINE.setBackground( Color.red );
    MINE.setOpaque( true );
  }
 
  /** The button displayed when the content of this cell is not revealed. */
  private final JButton btnButton= new JButton();

  /** The list of surrounders. */
  private final LinkedList lstSurrounders= new LinkedList();

  /** Indicates whether this cells contains a mine. */
  private boolean bContainsAMine= false;

  /** Indicates whether the player thinks that this cell contains a mine. */
  private boolean bPlayerSupposesAMine= false;

  /** Indicates whether the real content of this cell is revealed. */
  private boolean bIsRevealed= false;

  /** Stores the number of mines around this cell. */
  private int iSurroundingMines= 0;


  /**************************************** constructors ****************************************/

  /**
   * Creates a new mine field cell.
   */
  public MineFieldCell()
  {
  }


  /**************************************** getters ****************************************/

  /**
   * Returns the <code>Component</code> that is to be used to render this mine field cell,
   * or <code>null</code> if the cell is empty with no mines around.
   *
   * @post     return != null
   */
  public Component getComponent()
  {
    Component  cResult;                 // iContract requires only one return statement

    if (!isRevealed())
      /***** cell is not yet revealed *****/
      cResult= btnButton;
    else
      /***** cell is revealed *****/
      if (bContainsAMine)
	/***** mine ! *****/
        cResult= MINE;
      else
        /***** no mine *****/
 	if (bPlayerSupposesAMine)
          /***** player wrongly supposed a mine *****/
          cResult= ERROR;
        else
          /***** player made no assumption *****/
          if (getSurroundingMineCount() == 0)
            /***** no mine around *****/
            cResult= EMPTY;
          else
	    /***** mines around *****/
            cResult= DIGITS[getSurroundingMineCount()-1];
    return cResult;
  }

  /**
   * Indicates whether this cell has a mine.
   */
  public boolean doesContainAMine()
  {
    return bContainsAMine;
  }

  /**
   * Indicates the number of mines around this cell.
   *
   * @post     (0 <= return) && (return <= 8)
   */
  public int getSurroundingMineCount()
  {
    return iSurroundingMines;
  }

  /**
   * Indicates whether this cell has a flag.
   */
  public boolean isFlagged()
  {
    return bPlayerSupposesAMine;
  }

  /**
   * Indicates whether the real content of this cell is revealed.
   */
  public boolean isRevealed()
  {
    return bIsRevealed;
  }


  /**************************************** setters ****************************************/

  /**
   * Increase the number of mines around this cell.
   *
   * @pre      (getSurroundingMines() < 8)
   */
  public void increaseSurroundingMineCount()
  {
    ++iSurroundingMines;
  }

  /**
   * Registers the specified mine field cell as a surrounder of this one.
   */
  public void registerAsSurrounder( final MineFieldCell mfcSurrounder )
  {
    lstSurrounders.addLast( mfcSurrounder );
  }

  /**
   * Reset the mine field cell.
   */
  public void reset()
  {
    bContainsAMine= false;
    bPlayerSupposesAMine= false;
    bIsRevealed= false;
    iSurroundingMines= 0;
    btnButton.setIcon( null );
  }

  /**
   * Reveals the content of this cell.
   */
  public void reveal()
  {
    if (!bIsRevealed && !bPlayerSupposesAMine)
    {
      bIsRevealed= true;
      if (bContainsAMine)
        /***** mine ! *****/
        fireSteppedOnAMine();
      else
        if (iSurroundingMines == 0)
        {
          /***** no mine: recursive revelation *****/
          final ListIterator       liIterator= lstSurrounders.listIterator( 0 );

          while (liIterator.hasNext())
            ((MineFieldCell)liIterator.next()).reveal();
        }
      fireDisplayChanged();
    }
  }

  /**
   * Reveals the content of this cell if the user assumption was wrong.
   */
  public void revealError()
  {
    if (!bIsRevealed && bPlayerSupposesAMine && !bContainsAMine)
    {
      bIsRevealed= true;
      fireDisplayChanged();
    }
  }

  /**
   * Reveals the surrounders of this cell, if this cell is already revealed and the number
   * of surrounding flags matches the number of surrounding mines.
   */
  public void revealSurrounders()
  {
    if (bIsRevealed)
    {
      ListIterator       liIterator= lstSurrounders.listIterator( 0 );
      int                iSurroundingFlags= 0;

      while (liIterator.hasNext())
        if (((MineFieldCell)liIterator.next()).isFlagged())
          ++iSurroundingFlags;
      if (iSurroundingFlags == iSurroundingMines)
      {
        liIterator= lstSurrounders.listIterator( 0 );
        while (liIterator.hasNext())
          ((MineFieldCell)liIterator.next()).reveal();
      }
    }
  }

  /**
   * Settles a mine in this cell.
   *
   * @pre      !doesContainAMine()
   */
  public void settleAMine()
  {
    bContainsAMine= true;

    final ListIterator   liIterator= lstSurrounders.listIterator( 0 );

    while (liIterator.hasNext())
     ((MineFieldCell)liIterator.next()).increaseSurroundingMineCount();
  }

  /**
   * Toggles the flag.
   */
  public void toggleFlag()
  {
    if (!bIsRevealed)
    {
      bPlayerSupposesAMine= !bPlayerSupposesAMine;
      if (bPlayerSupposesAMine)
      {
        btnButton.setIcon( FLAG );
        fireFlagAdded();
      }
      else
      {
        btnButton.setIcon( null );
        fireFlagRemoved();
      }
      fireDisplayChanged();
    }
  }


  /**************************************** listener management ****************************************/

  /** The list of listeners. */
  private final EventListenerList lstListeners= new EventListenerList();

  /** The source of events. */
  private MineFieldCellEvent mfceEvent= null;

  /**
   * Adds the specified mine field cell listener to receive event from this cell.
   *
   * @pre      mfclListener != null
   */
  public void addListener( final MineFieldCellListener mfclListener )
  {
    lstListeners.add( MineFieldCellListener.class, mfclListener );
  }

  /**
   * Removes the specified mine field cell listener so that it no longer receive events from this cell.
   *
   * @pre      mfclListener != null
   */
  public void removeListener( final MineFieldCellListener mfclListener )
  {
    lstListeners.remove( MineFieldCellListener.class, mfclListener );
  }

  /**
   * Fires a displayChanged event.
   */
  protected void fireDisplayChanged()
  {
    final Object[]  oListeners= lstListeners.getListenerList();

    for ( int iListenerCounter= oListeners.length-2 ;
          iListenerCounter >= 0 ;
          iListenerCounter-= 2 )
      if (oListeners[iListenerCounter] == MineFieldCellListener.class)
      {
        if (mfceEvent == null)
          mfceEvent= new MineFieldCellEvent( this );
        ((MineFieldCellListener)oListeners[iListenerCounter+1]).displayChanged( mfceEvent );
      }
  }

  /**
   * Fires a flagAdded event.
   */
  protected void fireFlagAdded()
  {
    final Object[]  oListeners= lstListeners.getListenerList();

    for ( int iListenerCounter= oListeners.length-2 ;
          iListenerCounter >= 0 ;
          iListenerCounter-= 2 )
      if (oListeners[iListenerCounter] == MineFieldCellListener.class)
      {
        if (mfceEvent == null)
          mfceEvent= new MineFieldCellEvent( this );
        ((MineFieldCellListener)oListeners[iListenerCounter+1]).flagAdded( mfceEvent );
      }
  }

  /**
   * Fires a flagRemoved event.
   */
  protected void fireFlagRemoved()
  {
    final Object[]  oListeners= lstListeners.getListenerList();

    for ( int iListenerCounter= oListeners.length-2 ;
          iListenerCounter >= 0 ;
          iListenerCounter-= 2 )
      if (oListeners[iListenerCounter] == MineFieldCellListener.class)
      {
        if (mfceEvent == null)
          mfceEvent= new MineFieldCellEvent( this );
        ((MineFieldCellListener)oListeners[iListenerCounter+1]).flagRemoved( mfceEvent );
      }
  }

  /**
   * Fires a steppedOnAMine event.
   */
  protected void fireSteppedOnAMine()
  {
    final Object[]  oListeners= lstListeners.getListenerList();

    for ( int iListenerCounter= oListeners.length-2 ;
          iListenerCounter >= 0 ;
          iListenerCounter-= 2 )
      if (oListeners[iListenerCounter] == MineFieldCellListener.class)
      {
        if (mfceEvent == null)
          mfceEvent= new MineFieldCellEvent( this );
        ((MineFieldCellListener)oListeners[iListenerCounter+1]).steppedOnAMine( mfceEvent );
      }
  }
}
